#define G_LOG_DOMAIN "ap-ability"

#include "ap-ability.h"

#include <math.h>
#include <glib.h>
#include <glib-object.h>

typedef struct _APAbilityPrivate
{
    gchar *id;
    gchar *name;
    gint   score;
    gint   bonus;
    gint   modifier;
} APAbilityPrivate;

struct _APAbility
{
    GObject parent_instance;
};

/**
 * SECTION:ap-ability
 * @short_description: Manages a single ability of a character
 * @title: APAbility
 * @include: adventure-pack/ap-ability.h
 *
 * #APAbility is an object for creating and managing a character's ability.
 * Allows the setting of varouse ability attributes and automatically calculate
 * values like modifiers.
 **/

G_DEFINE_TYPE_WITH_PRIVATE (APAbility, ap_ability, G_TYPE_OBJECT)

enum
{
    PROP_0,
    PROP_ID,
    PROP_NAME,
    PROP_SCORE,
    PROP_BONUS,
    PROP_MODIFER,
    NUM_PROPERTIES
};

static GParamSpec *obj_properties[NUM_PROPERTIES] = { NULL, };

enum
{
    SCORE_CHANGED,
    N_SIGNALS
};

static guint signals[N_SIGNALS] = { 0, };

void
ap_ability_calculate_modifier (APAbility *self)
{
    APAbilityPrivate *priv = ap_ability_get_instance_private (self);
    gint score;

    /*
     * When a score number will result in a negative number (ex. 9 - 10 = -1)
     * then we subtract by 1 more to return a proper floor ().
     */
    if (priv->score < 10)
        score = priv->score - 1;
    else
        score = priv->score;

    /*
     * Modifer = ((Ability score) - 10) / 2
     */
    priv->modifier = floor ((score - 10) / 2);
}

static void
ap_ability_set_property (GObject      *object,
                         guint         property_id,
                         const GValue *value,
                         GParamSpec   *pspec)
{
    APAbility *self = AP_ABILITY (object);
    APAbilityPrivate *priv = ap_ability_get_instance_private (self);

    switch (property_id)
    {
        case PROP_ID:
            priv->id = g_value_dup_string (value);
            break;

        case PROP_NAME:
            priv->name = g_value_dup_string (value);
            break;

        case PROP_SCORE:
            priv->score = g_value_get_int (value);

            /* Recalculate modifier */
            ap_ability_calculate_modifier (self);

            g_signal_emit (self, signals[SCORE_CHANGED], 0, priv->id, priv->score); //TODO: With bonus?
            break;

        case PROP_BONUS:
            priv->bonus = g_value_get_int (value);

            //TODO: Propagate score-change when bonus changes
            break;

        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
            break;
    }
}

static void
ap_ability_get_property (GObject      *object,
                         guint         property_id,
                         GValue       *value,
                         GParamSpec   *pspec)
{
    APAbility *self = AP_ABILITY (object);
    APAbilityPrivate *priv = ap_ability_get_instance_private (self);

    switch (property_id)
    {
        case PROP_ID:
            g_value_set_string (value, priv->id);
            break;

        case PROP_NAME:
            g_value_set_string (value, priv->name);
            break;

        case PROP_SCORE:
            g_value_set_int (value, priv->score + priv->bonus);
            break;

        case PROP_BONUS:
            g_value_set_int (value, priv->bonus);
            break;

        case PROP_MODIFER:
            g_value_set_int (value, priv->modifier);
            break;

        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
            break;
    }
}

static void
ap_ability_init (APAbility *self)
{
}

static void
ap_ability_class_init (APAbilityClass *klass)
{
    GObjectClass *object_class = G_OBJECT_CLASS (klass);

    object_class->set_property = ap_ability_set_property;
    object_class->get_property = ap_ability_get_property;

	/**
	 * APAbility:id:
	 *
	 * ID of the ability, normally the ability name but no special characters,
     * lowercase and spaces replaced with dashes or underscores.
     *
     * ## Example
     * `Example Ability` becomes `example-ability`
	 */
    obj_properties[PROP_ID] =
        g_param_spec_string ("id",
                             "ID of the ability",
                             "ID of the ability, normally the ability name but no special characters, lowercase and spaces replaced with dashes or underscores.",
                             NULL, /* default value */
                             G_PARAM_CONSTRUCT_ONLY | G_PARAM_STATIC_NAME | G_PARAM_READWRITE);

	/**
	 * APAbility:name:
	 *
	 * Name of the ability.
	 */
    obj_properties[PROP_NAME] =
        g_param_spec_string ("name",
                             "Name of the ability",
                             "Name of the ability.",
                             NULL,  /* default value */
                             G_PARAM_READWRITE);

	/**
	 * APAbility:score:
	 *
	 * Score of the ability.
	 */
    obj_properties[PROP_SCORE] =
        g_param_spec_int ("score",
                          "Score of the ability",
                          "Score of the ability.",
                          G_MININT,
                          G_MAXINT,
                          0, /* default value */
                          G_PARAM_READWRITE);

	/**
	 * APAbility:bonus:
	 *
	 * Score bonus of the ability.
	 */
    obj_properties[PROP_BONUS] =
        g_param_spec_int ("bonus",
                          "Bonus score of the ability",
                          "Bonus score of the ability.",
                          G_MININT,
                          G_MAXINT,
                          0, /* default value */
                          G_PARAM_READWRITE);

	/**
	 * APAbility:modifer:
	 *
	 * (Read-only): Returns the ability modifier
     *
     * ((Ability Score) - 10) / 2
	 */
    obj_properties[PROP_MODIFER] =
        g_param_spec_int ("modifier",
                          "Modifier based of the ability score",
                          "Modifier based of the ability score.",
                          G_MININT,
                          G_MAXINT,
                          0, /* default value */
                          G_PARAM_READABLE);

    g_object_class_install_properties (object_class,
                                       NUM_PROPERTIES,
                                       obj_properties);

    signals[SCORE_CHANGED] =
        g_signal_new ("score-changed",
                      G_TYPE_FROM_CLASS (object_class),
                      G_SIGNAL_RUN_LAST,
                      0,
                      NULL,
                      NULL,
                      NULL,
                      G_TYPE_NONE,
                      2,
                      G_TYPE_STRING,
                      G_TYPE_INT);
}

/*
 * *****************************************************************************
 * API
 * *****************************************************************************
 */
/**
 * ap_ability_new:
 *
 * Creates a new #APAbility
 *
 * Returns: #APAbility
 */
APAbility *
ap_ability_new (gchar *id)
{
    return g_object_new (AP_TYPE_ABILITY,
                         "id", id,
                         NULL);
}

/**
 * ap_ability_get_id:
 * @self: An instance of #APAbility
 *
 * Gets the internal name (id) of the @self instance.
 *
 * Returns: (gchar *) (nullable): The internal name of @self (#APAbility)
 */
gchar *
ap_ability_get_id (APAbility *self)
{
    GValue val = G_VALUE_INIT;
    gchar *retval;

    g_assert (AP_IS_ABILITY (self));

    g_value_init (&val, G_TYPE_STRING);

    g_object_get_property (G_OBJECT (self), "id", &val);

    retval = g_value_dup_string (&val);

    g_value_unset (&val);

    return retval;
}

/**
 * ap_ability_set_name:
 * @self: An instance of #APAbility
 * @name: (gchar *): The ability name for the @self instance
 *
 * Sets the ability name for @self.
 *
 * ## Examples
 * * `Strength`
 * * `Dexterity`
 */
void
ap_ability_set_name (APAbility *self,
                     gchar     *name)
{
    GValue new_name = G_VALUE_INIT;

    g_assert (AP_IS_ABILITY (self));
    g_assert (name != NULL);
    g_assert (g_strcmp0 (name, "") != 0);

    g_value_init (&new_name, G_TYPE_STRING);
    g_value_set_string (&new_name, name);

    g_object_set_property (G_OBJECT (self), "name", &new_name);

    g_value_unset (&new_name);
}

/**
 * ap_ability_get_name:
 * @self: An instance of #APAbility
 *
 * Gets the name of the @self instance.
 *
 * Returns: (gchar *) (nullable): The ability name of @self (#APAbility)
 */
gchar *
ap_ability_get_name (APAbility *self)
{
    GValue val = G_VALUE_INIT;
    gchar *retval;

    g_assert (AP_IS_ABILITY (self));

    g_value_init (&val, G_TYPE_STRING);

    g_object_get_property (G_OBJECT (self), "name", &val);

    retval = g_value_dup_string (&val);

    g_value_unset (&val);

    return retval;
}

/**
 * ap_ability_set_score:
 * @self: An instance of #APAbility
 * @score: (gint): The ability score for the @self instance
 *
 * Sets the ability score for @self.
 */
void
ap_ability_set_score (APAbility *self,
                      gint       score)
{
    GValue new_score = G_VALUE_INIT;

    g_assert (AP_IS_ABILITY (self));

    g_value_init (&new_score, G_TYPE_INT);
    g_value_set_int (&new_score, score);

    g_object_set_property (G_OBJECT (self), "score", &new_score);

    g_value_unset (&new_score);
}

/**
 * ap_ability_get_score:
 * @self: An instance of #APAbility
 *
 * Gets the ability score of the @self instance.
 *
 * Returns: (gint) (nullable): The ability score of @self (#APAbility)
 */
gint
ap_ability_get_score (APAbility *self)
{
    GValue val = G_VALUE_INIT;
    gint retval;

    g_assert (AP_IS_ABILITY (self));

    g_value_init (&val, G_TYPE_INT);

    g_object_get_property (G_OBJECT (self), "score", &val);

    retval = g_value_get_int (&val);

    g_value_unset (&val);

    return retval;
}

/**
 * ap_ability_set_bonus:
 * @self: An instance of #APAbility
 * @bonus: (gint): The ability score bonus for the @self instance
 *
 * Sets the ability score bonus for @self.
 */
void
ap_ability_set_bonus (APAbility *self,
                      gint       bonus)
{
    GValue new_bonus = G_VALUE_INIT;

    g_assert (AP_IS_ABILITY (self));

    g_value_init (&new_bonus, G_TYPE_INT);
    g_value_set_int (&new_bonus, bonus);

    g_object_set_property (G_OBJECT (self), "bonus", &new_bonus);

    g_value_unset (&new_bonus);
}

/**
 * ap_ability_get_bonus:
 * @self: An instance of #APAbility
 *
 * Gets the ability score bonus of the @self instance.
 *
 * Returns: (gint) (nullable): The ability score bonus of @self (#APAbility)
 */
gint
ap_ability_get_bonus (APAbility *self)
{
    GValue val = G_VALUE_INIT;
    gint retval;

    g_assert (AP_IS_ABILITY (self));

    g_value_init (&val, G_TYPE_INT);

    g_object_get_property (G_OBJECT (self), "bonus", &val);

    retval = g_value_get_int (&val);

    g_value_unset (&val);

    return retval;
}

/**
 * ap_ability_get_modifier:
 * @self: An instance of #APAbility
 *
 * Gets the ability score modifier of the @self instance.
 *
 * Returns: (gint) (nullable): The ability score modifier of @self (#APAbility)
 */
gint
ap_ability_get_modifier (APAbility *self)
{
    GValue val = G_VALUE_INIT;
    gint retval;

    g_assert (AP_IS_ABILITY (self));

    g_value_init (&val, G_TYPE_INT);

    g_object_get_property (G_OBJECT (self), "modifier", &val);

    retval = g_value_get_int (&val);

    g_value_unset (&val);

    return retval;
}
